// HMethod.java, created Fri Jul 31 22:02:43 1998 by cananian
// Copyright (C) 1998 C. Scott Ananian <cananian@alumni.princeton.edu>
// Licensed under the terms of the GNU GPL; see COPYING for details.
package harpoon.ClassFile;

import harpoon.Util.ArrayFactory;

/**
 * An <code>HMethod</code> provides information about, and access to, a 
 * single method on a class or interface.  The reflected method
 * may be a class method or an instance method (including an abstract
 * method).
 * 
 * @author  C. Scott Ananian <cananian@alumni.princeton.edu>
 * @version $Id: HMethod.java,v 1.34 2003/03/18 03:52:53 salcianu Exp $
 * @see HMember
 * @see HClass
 */
public interface HMethod extends HMember {
  /**
   * Returns the <code>HClass</code> object representing the class or 
   * interface that declares the method represented by this
   * <code>HMethod</code> object. 
   */
  public HClass getDeclaringClass();

  /**
   * Returns the name of the method represented by this <code>HMethod</code>
   * object, as a <code>String</code>.
   */
  public String getName();

  /**
   * Returns the Java language modifiers for the method represented by this
   * <code>HMethod</code> object, as an integer.  The 
   * <code>java.lang.reflect.Modifier</code>
   * class should be used to decode the modifiers.
   * @see java.lang.reflect.Modifier
   */
  public int getModifiers();

  /**
   * Returns a <code>HClass</code> object that represents the formal
   * return type of the method represented by this <code>HMethod</code>
   * object.  Returns <code>HClass.Void</code> if the method returns
   * no value.
   */
  public HClass getReturnType();

  /**
   * Returns the descriptor for this method.  The format of a method
   * string descriptor is defined in <a
   * href="http://java.sun.com/docs/books/vmspec/html/ClassFile.doc.html#1169">Section
   * 4.3</a> of the JVM specification.  */
  public String getDescriptor();

  /**
   * Returns an array of <code>HClass</code> objects that represent the
   * formal parameter types, in declaration order, of the method
   * represented by this <code>HMethod</code> object.  Returns an array
   * of length 0 is the underlying method takes no parameters.
   */
  public HClass[] getParameterTypes();

  /**
   * Returns an array of <code>String</code> objects giving the declared
   * names of the formal parameters of the method.  The length of the
   * returned array is equal to the number of formal parameters.
   * If there is no <code>LocalVariableTable</code> attribute available
   * for this method, then every element of the returned array will be
   * <code>null</code>.
   */
  public String[] getParameterNames();

  /**
   * Returns an array of <code>HClass</code> objects that represent the
   * types of the checked exceptions thrown by the underlying method
   * represented by this <code>HMethod</code> object.  Returns an array
   * of length 0 if the method throws no checked exceptions.
   */
  public HClass[] getExceptionTypes();

  /**
   * Determines whether this <code>HMethod</code> is synthetic, 
   * in the sense of "associated with an inner-class; not programmer
   * declared".
   */
  public boolean isSynthetic();

  /** Determines whether this <code>HMethod</code> is an interface method.
   *  @return true if the declaring class is an interface, unless this
   *          method represents the static class initializer of the
   *          interface.
   */
  public boolean isInterfaceMethod();

  /** Determines whether this is a static method. */
  public boolean isStatic();

  /** Returns a mutator for this <code>HMethod</code>, or <code>null</code>
   *  if the object is immutable. */
  public HMethodMutator getMutator();
  
  /**
   * Compares this <code>HMethod</code> against the specified object.
   * Returns <code>true</code> if the objects are the same.  Two
   * <code>HMethod</code>s are the same if they were declared by the same
   * class and have the same name and formal parameter types.
   */
  public boolean equals(Object obj);

  /**
   * Returns a string describing this <code>HMethod</code>.  The string
   * is formatted as the method access modifiers, if any, followed by
   * the method return type, followed by a space, followed by the class
   * declaring the method, followed by a period, followed by the method
   * name, followed by a parenthesized, comma-separated list of the
   * method's formal parameter types.  If the method throws checked
   * exceptions, the parameter list is followed by a space, followed
   * by the word throws followed by a comma-separated list of the
   * throws exception types.  For example:<p>
   * <DL>
   * <DD><CODE>public boolean java.lang.Object.equals(java.lang.Object)</CODE>
   * </DL><p>
   * The access modifiers are placed in canonical order as specified by
   * "The Java Language Specification."  This is
   * <code>public</code>, <code>protected</code>, or <code>private</code>
   * first, and then other modifiers in the following order:
   * <code>abstract</code>, <code>static</code>, <code>final</code>, 
   * <code>synchronized</code>, <code>native</code>.
   */
  public String toString();

  /*****************************************************************/
  // JSR-14 extensions.

  /**
   * Returns an array of <code>HType</code> objects that represent the
   * formal parameter types, in declaration order, of the method
   * represented by this <code>HMethod</code> object.  Returns an
   * array of length 0 if the underlying method takes no parameters.
   * <p>
   * In particular, if the compile-time type of any formal parameter
   * is a type variable or a parameterized type, than an object of the
   * appropriate type (i.e., <code>HTypeVariable</code> or
   * <code>HParameterizedType</code>) will be returned.
   */
  public HType[] getGenericParameterTypes();
  /**
   * Returns an <code>HType</code> object that represents the formal
   * return type of the method represented by this
   * <code>HMethod</code> object.  In particular, if the compile-time
   * return type a parameterized type, than an object of the
   * appropriate type (i.e., <code>HTypeVariable</code> or
   * <code>HParameterizedType</code> will be returned.
   */
  public HType getGenericReturnType();
  /**
   * Returns an array of <code>HMethodTypeVariable</code> objects that
   * represent the type variables declared by the method represented
   * by this <code>HMethod</code> object, in declaration order.
   * Returns an array of length 0 if the underlying method declares no
   * type variables.
   */
  public HMethodTypeVariable[] getTypeParameters();

  /** Array factory: returns new <code>HMethod[]</code>. */
  public static final ArrayFactory<HMethod> arrayFactory =
    Factories.hmethodArrayFactory;
}

// set emacs indentation style.
// Local Variables:
// c-basic-offset:2
// End:
