// COWSetFactory.java, created Thu Jul 14 13:33:53 2005 by salcianu
// Copyright (C) 2005 Alexandru Salcianu <salcianu@alum.mit.edu>
// Licensed under the Modified BSD Licence; see COPYING for details.
package jpaul.DataStructs;

import java.util.Set;
import java.util.Collection;
import java.util.Iterator;

import jpaul.Misc.IntMCell;

/**
   <code>COWSetFactory</code> generates "copy-on-write" (COW) sets.  A
   COW set shares its representation (also a set) with other COW sets,
   until a mutation occurs.  At that moment, the COW set makes a
   private, exclusive copy of its underlying representation, and
   mutates that copy.

   <p>The internal representation of a COW set maintains a "sharing"
   counter to identify cases when the representation is not shared
   with anyone (and hence, no cloning is necessary before a mutation).

   <p>Cloning a COW set is a constant time operation.  COW sets are
   good when it is hard to determine statically whether a clone of a
   set will be mutated: they delay the real cloning until the first
   mutation (if any).

 * @author  Alexandru Salcianu - salcianu@alum.mit.edu
 * @version $Id: COWSetFactory.java,v 1.5 2005/08/18 15:39:51 salcianu Exp $ */
class COWSetFactory<T> extends SetFactory<T> {
    
    /** Creates a <code>COWSetFactory</code>.  

	@param underSetFact Set factory for generating the sets used
	in the representation of the COW sets generated by this
	<code>COWSetFactory</code>. */
    public COWSetFactory(SetFactory<T> underSetFact) {
        this.underSetFact = underSetFact;
    }

    private SetFactory<T> underSetFact;
    
    public Set<T> create() {
	return new COWSet<T>(underSetFact);
    }

    public Set<T> newColl(Collection<T> coll) {
	if(coll instanceof COWSet/*<T>*/) {
	    return (COWSet<T>) ((COWSet<T>) coll).clone();
	}
	return super.newColl(coll);
    }

    private static class SetWithCount<E2> {
	SetWithCount(Set<E2> set) {
	    this.set = set;
	}
	int countOthers = 0;
	final Set<E2> set;
    }


    private static class COWSet<E> implements Set<E>, Cloneable {

	COWSet(SetFactory<E> underSetFact) {
	    this.underSetFact = underSetFact;
	    this.underSet = new SetWithCount<E>(underSetFact.create());
	}

	private final SetFactory<E> underSetFact;
	private SetWithCount<E> underSet;

	public boolean add(E elem) {
	    if(underSet.countOthers != 0) {
		if(underSet.set.contains(elem)) {
		    return false;
		}
		else {
		    detach();
		}
	    }
	    return underSet.set.add(elem);
	}

	public boolean addAll(Collection<? extends E> c) {
	    if(c instanceof COWSet/*<E>*/) {
		COWSet cowSet2 = (COWSet<E>) c;
		// we just add a set to itself !
		if(this.underSet == cowSet2.underSet)
		    return false;
	    }
	    if(underSet.countOthers != 0) {
		if(underSet.set.containsAll(c)) {
		    return false;
		}
		else {
		    detach();
		}
	    }
	    return underSet.set.addAll(c);
	}
	
	public void clear() {
	    if(underSet.countOthers != 0) {
		if(underSet.set.isEmpty()) {
		    return;
		}
		else {
		    detach();
		}
	    }
	    underSet.set.clear();	    
	}

	public boolean contains(Object o) { 
	    return underSet.set.contains(o);
	}

	public boolean containsAll(Collection<?> c) {
	    return underSet.set.containsAll(c);
	}

	public boolean equals(Object o) {
	    if(o == null) return false;
	    if(o == this) return true;
	    if(o instanceof COWSet/*<E>*/) {
		COWSet<E> cowSet2 = (COWSet<E>) o;
		// hopefully, this is a common case :)
		if(this.underSet == cowSet2.underSet)
		    return true;
		return this.underSet.set.equals(cowSet2.underSet.set);
	    }
	    return this.underSet.set.equals(o);
	}

	public int     hashCode()  { return underSet.set.hashCode(); }
	public boolean isEmpty()   { return underSet.set.isEmpty(); }

	public boolean remove(Object o) {
	    if(underSet.countOthers != 0) {
		if(underSet.set.contains(o)) {
		    detach();
		}
		else {
		    return false;
		}
	    }
	    return underSet.set.remove(o);
	}

	public boolean removeAll(Collection<?> c) {
	    if(underSet.countOthers != 0) {
		detach();
	    }
	    return underSet.set.removeAll(c);
	}

	public boolean retainAll(Collection<?> c) {
	    throw new UnsupportedOperationException();
	}

	public int      size()         { return underSet.set.size(); }
	public Object[] toArray()      { return underSet.set.toArray(); }
	public <T> T[]  toArray(T[] a) { return underSet.set.toArray(a); }

	public Object clone() {
	    try {
		Object cloneObj = super.clone();
		synchronized(underSet) {
		    underSet.countOthers++;
		}
		return cloneObj;
	    }
	    catch(CloneNotSupportedException cex) {
		// should not happen
		throw new Error(cex);
	    }
	}

	public Iterator<E> iterator() {
	    return DSUtil.<E>unmodifiableIterator(underSet.set.iterator());
	}

	private void detach() {
	    SetWithCount<E> oldUnderSet = this.underSet;
	    this.underSet = new SetWithCount(underSetFact.create(oldUnderSet.set));
	    // do the decrement HERE, not before the real detach
	    // (it should limit the race conditions)
	    synchronized(oldUnderSet) {
		oldUnderSet.countOthers--;
	    }
	}

	public String toString() {
	    if(underSet.countOthers == 0) {
		return underSet.set.toString();
	    }
	    return 
		"(shared: " + (underSet.countOthers+1) + ") ; " +
		underSet.set.toString();
	}

	protected void finalize() {
	    try {
		try {
		    synchronized(underSet) {
			underSet.countOthers--;
		    }
		}
		finally {
		    super.finalize();
		}
	    }
	    catch(Throwable tex) {
		throw new Error(tex);
	    }
	}
    }
    
}
