// Relation.java, created Tue Jan 11 14:52:48 2000 by salcianu
// Copyright (C) 2000 Alexandru Salcianu - salcianu@alum.mit.edu
// Licensed under the Modified BSD Licence; see COPYING for details.
package jpaul.DataStructs;

import java.util.Collection;
import java.util.Collections;
import java.util.Set;
import java.util.Map;

import jpaul.Misc.Debug;
import jpaul.Misc.Predicate;

/**
 * <code>Relation</code> is a binary relation, accepting one to many
 * and many to one mappings.
 *
 * <p>It is similar to <code>net.cscott.jutil.MultiMap</code>.  One
 * difference: <code>Relation</code> does not implement the
 * <code>Map</code> interface.  I have always found this very
 * confusing; e.g., <code>MultiMap.get(v)</code> returns an arbitrary
 * value from the collection to which the multimap maps the value
 * <code>v</code>.
 *
 * @author  Alexandru Salcianu - salcianu@alum.mit.edu
 * @version $Id: Relation.java,v 1.13 2005/12/09 03:36:53 salcianu Exp $ */
public abstract class Relation<K,V> implements Cloneable {

    /** <code>Relation.EntryVisitor</code> is a wrapper for a function
	that is called on a relation entry of the form
	<code>&lt;key,value&gt;</code>.  Used by the method {@link
	#forAllEntries}.  */
    public static interface EntryVisitor<Key,Value> {
	/** Visits a <code>&lt;key,value&gt;</code> entry of a
	    relation.  May throw an
	    <code>InterruptTraversalException</code> in order to
	    indicate the desire to terminate the enclosing relation
	    traversal (the method {@link #forAllEntries}). */
	public void visit(Key key, Value value) throws InterruptTraversalException;
    }
    
    /** Adds the pair <code>&lt;key, value&gt;</code> to the relation.
	Returns <code>true</code> if the new relation is bigger. */
    public abstract boolean add(K key, V value);

    /** Adds a relation from <code>key</code> to each element of the set
	<code>values</code>. <code>values</code> should not contain
	duplicated elements.
	Returns <code>true</code> if the new relation is bigger. */
    public abstract boolean addAll(K key, Collection<V> values);


    /** Removes the relation between <code>key</code> and 
	<code>value</code>.
	@return <code>true</code> iff the relation changed */ 
    public abstract boolean remove(K key, V value);

    /** Removes the relation between <code>key</code> and 
	any element from <code>values</code>.
	@return <code>true</code> iff the relation changed */ 
    public abstract boolean removeAll(K key, Collection<V> values);


    /** Removes all the relations attached to <code>key</code>.
	@return <code>true</code> iff the relation changed */    
    public abstract boolean removeKey(K key);

    /** Removes all the keys that satisfy <code>predicate.check()</code>.
	@return <code>true</code> iff the relation changed */
    public abstract boolean removeKeys(Predicate<K> predicate);


    /** Removes all the values that satisfy <code>predicate.check()</code>.
	@return <code>true</code> iff the relation changed */
    public abstract boolean removeValues(Predicate<V> predicate);


    /** Checks the existence of the relation <code>&lt;key,value&gt;</code>. */
    public boolean contains(K key, V value) {
	return _getValues(key).contains(value);
    }

    /** Checks the existence of the relation between <code>key</code>
        and every element from <code>values</code>. */
    public boolean containsAll(K key, Collection<V> values) {
	return _getValues(key).containsAll(values);
    }

    /** Checks the existence of the <code>key</code> key in this relation. */
    public abstract boolean containsKey(K key);


    /** Tests if this relation is empty or not. */
    public abstract boolean isEmpty();


    /** Returns the image of <code>key</code> through this relation.
	The returned collection IS IMMUTABLE.
	Returns <code>Collections.emptySet()</code> if no value is attached to <code>key</code>. */
    public final Set<V> getValues(K key) {
	Set<V> set = _getValues(key);
	if(set == null)
	    return Collections.emptySet();
	else 
	    return Collections.unmodifiableSet(set);
    }

    /** Method used by the internal implementation of the
        <code>Relation</code> or its subclasses.  Similar to the
        user-level {@link #getValues} but the returned set IS MUTABLE.
        Mutating this set affects the set of values that are
        associated with a given key; therefore, this method cannot be
        used by <code>Relation</code> clients directly. */
    protected abstract Set<V> _getValues(K key);


    /** Returns an IMMUTABLE view of all the keys appearing in
	<code>this</code> relation.  If you want to delete a key, then
	use {@link #removeKey}.  */
    public abstract Set<K> keys();


    /** Returns an IMMUTABLE view of all the values appearing in
        <code>this</code> relation.  The view may contain the same
        value twice, if it is associated with two distinct keys.  */
    public abstract Iterable<V> values();


    /** Combines <code>this</code> relation with relation
	<code>rel</code>.  A <code>null</code> parameter is considered
	to be an empty relation.

	@return <code>true</code> iff <code>this</code> relation has
	changed. */
    public abstract boolean union(Relation<K,V> rel);


    /** Checks the equality of two relations */
    public abstract boolean equals(Object o);


    /** Visits all the entries <code>&lt;key,value&gt;</code> of
	<code>this</code> relation and calls
	<code>visitor.visit</code> on each of them.  This traversal of
	the relation entries can be stopped at any point by throwing
	an {@link jpaul.DataStructs.InterruptTraversalException
	InterruptTraversalException} from the visitor; the exception
	is caught internally by the implementation of this method. */
    public void forAllEntries(EntryVisitor<K,V> visitor) {
	try {
	    for(K key : keys()) {
		for(V value : _getValues(key)) {
		    visitor.visit(key, value);
		}
	    }
	}
	catch(InterruptTraversalException itex) {
	    // Do nothing; InterruptTraversalException is only a way
	    // to terminate the traversal prematurely.
	}
    }


    /** Revert <code>this</code> relation and store the result into
	the relation <code>result</code>. &lt;a,b&gt; appears in the
	reverse relation iff &lt;b,a&gt; appears in <code>this</code>
	relation.  Returns the new relation (ie,
	<code>result</code>). */
    public Relation<V,K> revert(final Relation<V,K> result) {
	forAllEntries(new EntryVisitor<K,V>() {
		public void visit(K key, V value) {
		    result.add(value, key);
		}
	    });
	return result;
    }


    /** Returns the number of &lt;key,value&gt; pairs in
        <code>this</code> relation.  Linear in the size of the
        relation.  This may be also implemented in O(1) by
        incrementally updating a <code>size</code> field, but that may
        complicate the code in the presence of subclassing, etc.  Will
        think about it if it becomes a problem. */
    public int size() {
	int size = 0;
	for(K key : keys()) {
	    size += getValues(key).size();
	}
	return size;
    }


    public Object clone() {
	try {
	    return super.clone();
	}
	catch(CloneNotSupportedException e) {
	    // should not happen ...
	    throw new Error(e);
	}
    }


    /** Pretty-print function for debug.
	<code>rel1.equals(rel2) <==> rel1.toString().equals(rel2.toString())</code> */
    public String toString() {
	StringBuffer buffer = new StringBuffer();

	buffer.append("{");
	
	for(K key : Debug.sortedCollection(keys())) {
	    buffer.append("\n  ");		
	    buffer.append(key);
	    buffer.append(" -> ");
	    buffer.append(Debug.stringImg(_getValues(key)));
	}
	
	buffer.append("\n }\n");
	
	return buffer.toString();
    }

    // THE FOLLOWING FEATURES WOULD BE NICE TO ADD IN THE FUTURE

    /* Removes all the relations involving at least one object that
	satisfy <code>predicate.check()</code>. */
    //public void removeObjects(Predicate predicate);


    /* Returns the subrelation of this relation that contains
	only the keys that appear in <code>selected_keys</code>. */
    //public Relation<K,Value> select(Collection<K> selected_keys);


    /* Convert <code>this</code> relation through the mapping
	<code>map</code>.  The converted mapping contains all pairs
	<code>(a,b)</code> such that there exists <code>c,d</code>
	such that <code>(c,d)</code> appears in <code>this</code>
	mapping, and map maps c to a and d to b.  If an object is not
	mapped to anything by <code>map</code>, it will be mapped to
	itself by default.  The result is stored in
	<code>result</code>.  Returns the converted mapping (ie,
	result).  */
    //public<K2,V2> Relation<K2,V2> convert(Map<K,K2> keyMap, Map<V,V2> valueMap, Relation<K2,V2> result);


    public static <K,V> Relation<K,V> unmodifiableRelation(final Relation<K,V> rel) {
	return new Relation<K,V>() {
	    public boolean add(K key, V value) {
		throw new UnsupportedOperationException("unmodifiable Relation");
	    }
	    public boolean addAll(K key, Collection<V> values) {
		throw new UnsupportedOperationException("unmodifiable Relation");
	    }
	    public boolean remove(K key, V value) {
		throw new UnsupportedOperationException("unmodifiable Relation");
	    }
	    public boolean removeAll(K key, Collection<V> values) {
		throw new UnsupportedOperationException("unmodifiable Relation");
	    }
	    public boolean removeKey(K key) {
		throw new UnsupportedOperationException("unmodifiable Relation");
	    }
	    public boolean removeKeys(Predicate<K> predicate) {
		throw new UnsupportedOperationException("unmodifiable Relation");
	    }
	    public boolean removeValues(Predicate<V> predicate) {
		throw new UnsupportedOperationException("unmodifiable Relation");
	    }
	    public boolean contains(K key, V value) {
		return rel.contains(key, value);
	    }
	    public boolean containsKey(K key) {
		return rel.containsKey(key);
	    }
	    public boolean isEmpty() { return rel.isEmpty(); }
	    // No need to put an immutable wrapper around the returned
	    // set.  This method is used only internally; moreover, it
	    // can be mutated only by methods that mutate the overall
	    // relation, and these methods are already forbidden.
	    protected Set<V> _getValues(K key) { return rel._getValues(key); }
	    // we don't wrap rel.keys() in an immutable wrapper:
	    // rel.keys() is supposed to be immutable anyway.
	    public Set<K> keys() { return rel.keys(); }
	    // same immutability considerations as for keys()
	    public Iterable<V> values() { return rel.values(); }
	    public boolean union(Relation<K,V> rel) {
		throw new UnsupportedOperationException("unmodifiable Relation");
	    }
	    public boolean equals(Object o) { return rel.equals(o); }
	    public void forAllEntries(EntryVisitor<K,V> visitor) {
		rel.forAllEntries(visitor);
	    }
	    public Object clone() { return super.clone(); }
	};
    }

}
